/*
 * Copyright (c) 1988, Julian Onions <jpo@cs.nott.ac.uk>
 * Nottingham University 1987.
 *
 * This source may be freely distributed, however I would be interested
 * in any changes that are made.
 *
 * This driver takes packets off the IP i/f and hands them up to a
 * user process to have it's wicked way with. This driver has it's
 * roots in a similar driver written by Phil Cockcroft (formerly) at
 * UCL. This driver is based much more on read/write/select mode of
 * operation though.
 * 
 * $Id: if_tun.c,v 1.9 1993/12/24 03:20:59 deraadt Exp $
 */
#define	TUN_MAJOR	28

#define	NTUN	4
#define	INET	1
#if NTUN > 0

#define	KERNEL	1
#define	KERNEL_FEATURES	1

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/conf.h>

#if 0
#include <machine/cpu.h>
#endif
#include <kernserv/machine/spl.h>

#include <kernserv/kern_server_types.h>

#include <net/if.h>
#include <net/netif.h>
#include <net/netisr.h>
#include <net/route.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#include "if_tun.h"

#define TUNDEBUG	if (tundebug) printf
int	tundebug = 0;

kern_server_t instance;

struct tun_softc tunctl[NTUN];
extern int ifqmaxlen;

int tunoutput();
netbuf_t tungetbuf(netif_t);
int	tuncontrol(netif_t, const char *, void *);

/*
	Functions which enqueue, dequeue and check queue of netbuf.
	Derived from if_ppp.c (PPP0.3 written by Drew D. Perkins)
*/
static inline int
nbq_full(struct nb_queue* nbq)
{
    if (nbq->head == 0)
	    return nbq->tail == QUEUE_SIZE - 1;
    else
	    return nbq->head - 1 == nbq->tail;
}

static inline int
nbq_empty(struct nb_queue* nbq)
{
    return nbq->head == nbq->tail;
}

static inline netbuf_t
nbq_dequeue(struct nb_queue* nbq)
{
    int s = splimp();
    if (nbq_empty(nbq))
    {
	return NULL;
    }
    else
    {
	netbuf_t ret = nbq->queue[nbq->head];
	if (nbq->head == QUEUE_SIZE - 1)
		nbq->head = 0;
	else
		nbq->head++;
	return ret;
    }
    splx(s);
}

static inline int
nbq_enqueue(struct nb_queue* nbq, netbuf_t nb)
{
    int s = splimp();
    if (nbq_full(nbq))
    {
	return -1;
    }
    else
    {
	nbq->queue[nbq->tail] = nb;
	if (nbq->tail == QUEUE_SIZE - 1)
		nbq->tail = 0;
	else
		nbq->tail++;
    }
    splx(s);
	return 0;	/* Feb.11,1994 K.Suzuki */
}

/*
 * get buffer - return newly allocated netbuffer
 */
netbuf_t
tungetbuf(netif_t ifp)
{
    struct ifnet* ifn = (struct ifnet*)ifp;
    int len = MAX(ifn->if_mtu, TUNMTU) + sizeof (struct ifnet *);
    return nb_alloc(len);
}

/*
 * tunnel open - must be superuser & the device must be
 * configured in
 */
int
tunopen(dev, flag, mode, p)
	dev_t	dev;
	int	flag, mode;
	struct proc *p;
{
	struct ifnet	*ifp;
	struct tun_softc *tp;
	register int	unit, error;

#if 0
	if (error = suser(p->p_ucred, &p->p_acflag))
		return (error);
#endif

	if ((unit = minor(dev)) >= NTUN)
		return (ENXIO);
	tp = &tunctl[unit];
	if (tp->tun_flags & TUN_OPEN)
		return ENXIO;
	ifp = tp->tun_if;
	tp->tun_flags |= TUN_OPEN;
	TUNDEBUG("%s%d: open\n", ifp->if_name, ifp->if_unit);
	return (0);
}

/*
 * tunclose - close the device - mark i/f down & delete
 * routing info
 */
int
tunclose(dev, flag)
	dev_t	dev;
	int	flag;
{
	register int	unit = minor(dev), s;
	struct tun_softc *tp = &tunctl[unit];
	struct ifnet	*ifp = tp->tun_if;
	netbuf_t	nb;

	tp->tun_flags &= ~TUN_OPEN;

	/*
	 * junk all pending output
	 */
	do {
		s = splimp();
		nb = nbq_dequeue( &tp->if_snd);
		splx(s);
		if( nb)
			nb_free( nb);
	} while (nb);

	if (ifp->if_flags & IFF_UP) {
		s = splimp();
		if_down(ifp);
		if (ifp->if_flags & IFF_RUNNING) {
#ifdef NeXT
#else
			rtinit(ifp->if_addrlist, (int)RTM_DELETE,
			       tp->tun_flags & TUN_DSTADDR ? RTF_HOST : 0);
#endif
		}
		splx(s);
	}
	tp->tun_pgrp = 0;
	if( tp->tun_rsel)
		selthreadclear( &tp->tun_rsel);

	TUNDEBUG ("%s%d: closed\n", ifp->if_name, ifp->if_unit);
	return (0);
}

/*
 * tuninit - initialize the network I/F
 *
 */
static int
tuninit(unit)
	int	unit;
{
	struct tun_softc *tp = &tunctl[unit];
	struct ifnet	*ifp = tp->tun_if;
	register struct ifaddr *ifa;

	TUNDEBUG("%s%d: tuninit\n", ifp->if_name, ifp->if_unit);

	if_flags_set( (netif_t)ifp, if_flags((netif_t)ifp)|(IFF_UP|IFF_RUNNING));

	for (ifa = ifp->if_addrlist; ifa; ifa = ifa->ifa_next) {
		struct sockaddr_in *si;

		si = (struct sockaddr_in *)&ifa->ifa_addr;
		if (si && si->sin_addr.s_addr)
			tp->tun_flags |= TUN_IASET;

		si = (struct sockaddr_in *)&ifa->ifa_dstaddr;
		if (si && si->sin_addr.s_addr)
			tp->tun_flags |= TUN_DSTADDR;
	}

	return 0;
}

/*
 * process ioctl request
 */
int
tuncontrol(ifp, cmd, data)
netif_t ifp;
const char* cmd;
void* data;
{
    if (!strcmp(cmd, "unix-ioctl"))
    {
	if_ioctl_t* ctl = (if_ioctl_t*)data;
	return tunifioctl((struct ifnet*)ifp,
			ctl->ioctl_command,
			ctl->ioctl_data);
    }
    else if (!strcmp(cmd, "setaddr"))
    {
	struct sockaddr_in *sin = (struct sockaddr_in *)data;
	if (sin->sin_family != AF_INET)
		return EAFNOSUPPORT;
	tuninit( ((struct ifnet *)ifp)->if_unit);
	return 0;
    }
    else if (!strcmp(cmd, "setflags"))
    {
	return 0;
    }
    else
    {
	printf("Invalid tun control %s\n", cmd);
	return EINVAL;
    }
}

/*
 * Process an unix-ioctl request.
 */
int
tunifioctl(ifp, cmd, data, flag)
	struct ifnet *ifp;
	int	cmd;
	caddr_t	data;
	int	flag;
{
	struct tun_softc *tp = &tunctl[ifp->if_unit];
	int		error = 0, s;

	s = splimp();
	switch(cmd) {
	case SIOCSIFADDR:
		tuninit(ifp->if_unit);
		break;
	case SIOCSIFDSTADDR:
		tp->tun_flags |= TUN_DSTADDR;
		TUNDEBUG("%s%d: destination address set\n", ifp->if_name,
		    ifp->if_unit);
		break;
	default:
		error = EINVAL;
	}
	splx(s);
	return (error);
}

/*
 * tunoutput - queue packets from higher level ready to put out.
 */
int
tunoutput(ifp, nb, dst)
	struct ifnet    *ifp;
	netbuf_t        nb;
	struct sockaddr *dst;
{
	struct tun_softc *tp = &tunctl[ifp->if_unit];
	struct proc	*p;
	int		s;

	TUNDEBUG ("%s%d: tunoutput\n", ifp->if_name, ifp->if_unit);

	if ((tp->tun_flags & TUN_READY) != TUN_READY) {
		TUNDEBUG ("%s%d: not ready 0%o\n", ifp->if_name,
			  ifp->if_unit, tp->tun_flags);
		nb_free(nb);
		return EHOSTDOWN;
	}

	switch(dst->sa_family) {
#ifdef INET
	case AF_INET:
		s = splimp();
		if( nbq_enqueue( &tp->if_snd, nb) < 0) {
			nb_free(nb);
			splx(s);
			ifp->if_collisions++;
			return (ENOBUFS);
		}
		splx(s);
		ifp->if_opackets++;
		break;
#endif
	default:
		TUNDEBUG ("Unkown protocol: tunoutput\n");
		nb_free(nb);
		return EAFNOSUPPORT;
	}

	if (tp->tun_flags & TUN_RWAIT) {
		tp->tun_flags &= ~TUN_RWAIT;
		wakeup((caddr_t)tp);
	}
	if (tp->tun_flags & TUN_ASYNC && tp->tun_pgrp) {
		if (tp->tun_pgrp > 0)
			gsignal(tp->tun_pgrp, SIGIO);
		else if (p = pfind(-tp->tun_pgrp))
			psignal(p, SIGIO);
	}
	if (tp->tun_rsel) {
		selwakeup(tp->tun_rsel, tp->tun_flags & TUN_RCOLL);
		tp->tun_flags &= ~TUN_RCOLL;
		selthreadclear( &tp->tun_rsel);
	}
	return 0;
}

/*
 * the cdevsw interface is now pretty minimal.
 */
int
tunioctl(dev, cmd, data, flag)
	dev_t		dev;
	int		cmd;
	caddr_t		data;
	int		flag;
{
	int		unit = minor(dev), s;
	struct tun_softc *tp = &tunctl[unit];
	struct tuninfo *tunp;

	switch (cmd) {
	case TUNSIFINFO:
	        tunp = (struct tuninfo *)data;
		tp->tun_if->if_mtu = tunp->if_mtu;
		strcpy( tp->tun_if->if_type, tunp->if_type);
		tp->if_baudrate = tunp->if_baudrate;
		break;
	case TUNGIFINFO:
		tunp = (struct tuninfo *)data;
		tunp->if_mtu = tp->tun_if->if_mtu;
		strcpy( tunp->if_type, tp->tun_if->if_type);
		tunp->if_baudrate = tp->if_baudrate;
		break;
	case TUNSDEBUG:
		tundebug = *(int *)data;
		break;
	case TUNGDEBUG:
		*(int *)data = tundebug;
		break;
	case FIONBIO:
		if (*(int *)data)
			tp->tun_flags |= TUN_NBIO;
		else
			tp->tun_flags &= ~TUN_NBIO;
		break;
	case FIOASYNC:
		if (*(int *)data)
			tp->tun_flags |= TUN_ASYNC;
		else
			tp->tun_flags &= ~TUN_ASYNC;
		break;
	case FIONREAD:
		s = splimp();
		if( !nbq_empty( &tp->if_snd))
			*(int *)data = nb_size( tp->if_snd.queue[ tp->if_snd.head]);
		else	
			*(int *)data = 0;
		splx(s);
		break;
	case TIOCSPGRP:
		tp->tun_pgrp = *(int *)data;
		break;
	case TIOCGPGRP:
		*(int *)data = tp->tun_pgrp;
		break;
	default:
		return (ENOTTY);
	}
	return (0);
}

/*
 * The cdevsw read interface - reads a packet at a time, or at
 * least as much of a packet as can be read.
 */
int
tunread(dev, uio)
	dev_t		dev;
	struct uio	*uio;
{
	int		unit = minor(dev);
	struct tun_softc *tp = &tunctl[unit];
	struct ifnet	*ifp = tp->tun_if;
	netbuf_t	nb;
	int		error=0, len, s;

	TUNDEBUG ("%s%d: read\n", ifp->if_name, ifp->if_unit);
	if ((tp->tun_flags & TUN_READY) != TUN_READY) {
		TUNDEBUG ("%s%d: not ready 0%o\n", ifp->if_name,
			  ifp->if_unit, tp->tun_flags);
		return EHOSTDOWN;
	}

	tp->tun_flags &= ~TUN_RWAIT;

	s = splimp();
	do {
		nb = nbq_dequeue( &tp->if_snd);
		if (nb == 0) {
			if (tp->tun_flags & TUN_NBIO) {
				splx(s);
				return EWOULDBLOCK;
			}
			tp->tun_flags |= TUN_RWAIT;
			sleep((caddr_t)tp, PZERO + 1);
		}
	} while (nb == 0);
	splx(s);

	while (nb && uio->uio_resid > 0 && error == 0) {
		len = MIN(uio->uio_resid, nb_size(nb));
		if (len == 0)
			break;
		error = uiomove(nb_map(nb), len, UIO_READ, uio);
		nb_free(nb);
		nb = 0;
	}

	if (nb) {
		TUNDEBUG("Dropping netbuf\n");
		nb_free(nb);
	}
	return error;
}

/*
 * the cdevsw write interface - an atomic write is a packet - or else!
 */
int
tunwrite(dev, uio)
	dev_t		dev;
	struct uio	*uio;
{
	int		unit = minor (dev);
	struct ifnet	*ifp = tunctl[unit].tun_if;
	netbuf_t	nb;
	int		error=0, len;

	TUNDEBUG("%s%d: tunwrite\n", ifp->if_name, ifp->if_unit);

	if (uio->uio_resid < 0 || uio->uio_resid > TUNMTU) {
		TUNDEBUG("%s%d: len=%d!\n", ifp->if_name, ifp->if_unit,
		    uio->uio_resid);
		return EIO;
	}

	nb = tungetbuf( (netif_t)ifp);
	len = uio->uio_resid;
	
	if (error = uiomove(nb_map(nb), nb_size(nb), UIO_WRITE, uio))
	{
	    nb_free(nb);
	    return error;
	}
	nb_shrink_bot(nb, nb_size(nb) - len);
	inet_queue( ifp, nb);
	ifp->if_ipackets++;
	return error;
}

/*
 * tunselect - the select interface, this is only useful on reads
 * really. The write detect always returns true, write never blocks
 * anyway, it either accepts the packet or drops it.
 */
int
tunselect(dev, rw)
	dev_t		dev;
	int		rw;
{
	int		unit = minor(dev), s;
	/* struct proc		*p = u.u_procp; */
	struct tun_softc *tp = &tunctl[unit];
	struct ifnet	*ifp = tp->tun_if;

	s = splimp();
	TUNDEBUG("%s%d: tunselect\n", ifp->if_name, ifp->if_unit);

	switch (rw) {
	case FREAD:
		if( !nbq_empty(&tp->if_snd)) {
			splx(s);
			TUNDEBUG("%s%d: tunselect", ifp->if_name,
			    ifp->if_unit);
			return 1;
		}
		if( selthreadcache( &tp->tun_rsel))
			tp->tun_flags |= TUN_RCOLL;
		break;
	case FWRITE:
		splx(s);
		return 1;
	}
	splx(s);
	TUNDEBUG("%s%d: tunselect waiting\n", ifp->if_name, ifp->if_unit);
	return 0;
}

/*
 * attach the tunnel device to os
 */

struct cdevsw	tun_cdevsw = {
	tunopen,
	tunclose,
	tunread,
	tunwrite,
	tunioctl,
	nodev,
	nulldev,
	tunselect,
	nodev,
	nodev,
	nodev};

struct cdevsw	saved_cdevsw;

void
tunattach(unused)
	int unused;
{
	register int i;
	struct ifnet *ifp;
	struct sockaddr_in *sin;
	extern int ipforwarding;
	extern int ipsendredirects;

	for (i = 0; i < NTUN; i++) {
		tunctl[i].tun_flags = TUN_INITED;

		tunctl[i].tun_if = (struct ifnet*)
			if_attach(NULL, NULL, tunoutput, 
				  tungetbuf, tuncontrol, "tun", i, "Tunnel Device", 
				  TUNMTU, IFF_POINTOPOINT, NETIFCLASS_VIRTUAL, NULL);
	}
	ipforwarding = 1;
	ipsendredirects = 1;

	saved_cdevsw = cdevsw[ TUN_MAJOR];
	cdevsw[ TUN_MAJOR] = tun_cdevsw;

	printf( "Tunnel Device for NS3.2J, version 0.1 by Kazuhiko Suzuki\n");
}

/*
 * Detach the tunnel device
 */
void
tundetach( void)
{
	int		i;
	for( i=0; i<NTUN; i++)
		if_detach( (netif_t)tunctl[i].tun_if);
	cdevsw[ TUN_MAJOR] = saved_cdevsw;

	printf( "Remove Tunnel Device\n");
}


#endif  /* NTUN */
